import logging
import os
import sys
from logging.handlers import RotatingFileHandler

try:
    from flask import current_app, has_request_context, request
except ImportError:
    current_app = None
    def has_request_context(): return False
    request = None


class RequestFormatter(logging.Formatter):
    def format(self, record):
        if has_request_context():
            record.url = getattr(request, "url", None)
            record.remote_addr = getattr(request, "remote_addr", None)
            record.method = getattr(request, "method", None)
        else:
            record.url = None
            record.remote_addr = None
            record.method = None
        return super().format(record)


LOG_FORMAT = (
    "[%(asctime)s] %(levelname)s in %(module)s [%(process)d]: %(message)s"
    " | url=%(url)s remote_addr=%(remote_addr)s method=%(method)s"
)


def get_logger(name="app", config=None):
    logger = logging.getLogger(name)

    # Bersihkan handler sebelumnya jika ada
    if logger.hasHandlers():
        logger.handlers.clear()

    # Ambil konfigurasi dari Flask config atau fallback ke environment variable
    if config is None:
        try:
            config = current_app.config
        except Exception:
            config = {}

    LOG_LEVEL = str(config.get("LOG_LEVEL", os.getenv("LOG_LEVEL", "INFO"))).upper()
    LOG_DIR = config.get("LOG_DIR", os.getenv("LOG_DIR", "logs"))
    LOG_FILE = config.get("LOG_FILE", os.getenv("LOG_FILE", "app.log"))
    LOG_PATH = os.path.join(LOG_DIR, LOG_FILE)
    LOG_MAX_BYTES = int(config.get("LOG_MAX_BYTES", os.getenv("LOG_MAX_BYTES", 10 * 1024 * 1024)))
    LOG_BACKUP_COUNT = int(config.get("LOG_BACKUP_COUNT", os.getenv("LOG_BACKUP_COUNT", 5)))

    os.makedirs(LOG_DIR, exist_ok=True)

    formatter = RequestFormatter(LOG_FORMAT)

    # File log handler (jika bisa)
    try:
        file_handler = RotatingFileHandler(
            LOG_PATH, maxBytes=LOG_MAX_BYTES, backupCount=LOG_BACKUP_COUNT, encoding="utf-8"
        )
        file_handler.setLevel(LOG_LEVEL)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    except Exception as e:
        print(f"[LOGGER INIT] Could not add file handler: {e}")

    # Console stream handler yang aman
    try:
        console_handler = logging.StreamHandler(sys.__stdout__)  # lebih aman daripada sys.stdout.buffer
        console_handler.setLevel(LOG_LEVEL)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    except Exception as e:
        print(f"[LOGGER INIT] Could not add console handler: {e}")

    logger.setLevel(LOG_LEVEL)
    logger.propagate = False

    return logger



import json
import uuid

def build_gateway_map(elements):
    mapping = {}
    for gw in elements.get("gateways", []):
        gw_id = gw.get("id")
        gw_type = (gw.get("gateway_type") or "").lower().strip()
        gw_name = (gw.get("name") or "").lower()

        if not gw_type:
            if "event" in gw_name:
                gw_type = "eventbasedgateway"
            elif "complex" in gw_name:
                gw_type = "complexgateway"
            elif "exclusive" in gw_name:
                gw_type = "exclusivegateway"
            elif "parallel" in gw_name:
                gw_type = "parallelgateway"
            elif "inclusive" in gw_name:
                gw_type = "inclusivegateway"
            else:
                gw_type = "gateway"
        mapping[gw_id] = gw_type
    return mapping


def classify_gateway_direction(gateway_id, incoming, outgoing, gateway_type=None):
    if gateway_type and gateway_type.lower() == "eventbasedgateway":
        return "SPLIT"

    in_count = len(incoming.get(gateway_id, []))
    out_count = len(outgoing.get(gateway_id, []))

    if out_count > 1:
        return "SPLIT"
    elif in_count > 1:
        return "JOIN"
    else:
        return "SINGLE"


def gateway_label(base_type, direction):
    mapping = {
        "exclusivegateway": "XOR",
        "parallelgateway": "AND",
        "inclusivegateway": "OR",
        "complexgateway": "COMPLEX",
        "eventbasedgateway": "EVENT_BASED",
    }

    prefix = mapping.get(base_type.lower(), "GATEWAY")

    if base_type.lower() == "eventbasedgateway":
        direction = "SPLIT"

    if direction in ["SPLIT", "JOIN"]:
        return f"{prefix}_{direction}"
    return prefix

import json
import uuid

# ... pastikan fungsi build_gateway_map, classify_gateway_direction, gateway_label ada di file ...

def generate_edges(elements, process_id=None):
    cypher = []
    gateway_map = build_gateway_map(elements)

    def to_cypher_value(val):
        return 'null' if val is None else json.dumps(val)

    # ambil semua flows awal (message flows juga)
    original_flows = elements.get("flows", []) + elements.get("flows_by_type", {}).get("message_flows", [])

    # --- DETEKSI gateway -> gateway SAAT MASIH BERBENTUK ASLI (hanya sekali) ---
    gateway_chains = set()
    for f in original_flows:
        s, t = f.get("source"), f.get("target")
        if s in gateway_map and t in gateway_map:
            gateway_chains.add(s)
            gateway_chains.add(t)

    # --- buat invisible tasks bila ada gateway->gateway ---
    invisible_tasks = []
    new_flows = []

    for f in list(original_flows):
        src, tgt = f.get("source"), f.get("target")
        if src in gateway_map and tgt in gateway_map:
            # buat invisible task node unik
            invisible_id = f"invisible_{uuid.uuid4().hex[:8]}"
            src_flow_info = next((x for x in original_flows if x.get("source") == src), None)
            prev_node = {
                "pool_id": src_flow_info.get("source_pool") if src_flow_info else None,
                "lane_id": src_flow_info.get("source_lane") if src_flow_info else None,
                "pool_name": src_flow_info.get("source_pool_name") if src_flow_info else None,
                "lane_name": src_flow_info.get("source_lane_name") if src_flow_info else None,
            }

            invisible_tasks.append({
                "id": invisible_id,
                "name": "Invisible Task",
                "type": "InvisibleTask",
                "prev_node": prev_node
            })

            f1 = dict(f)
            f1["id"] = f"{f['id']}_inv_in"
            f1["target"] = invisible_id

            f2 = dict(f)
            f2["id"] = f"{f['id']}_inv_out"
            f2["source"] = invisible_id

            new_flows.extend([f1, f2])
        else:
            new_flows.append(f)

    # gunakan new_flows sebagai all_flows selanjutnya (sudah termasuk invisible flows)
    all_flows = new_flows

    # sekarang bangun incoming/outgoing berdasarkan all_flows yang sudah disesuaikan
    incoming, outgoing = {}, {}
    for flow in all_flows:
        src = flow.get("source")
        tgt = flow.get("target")
        outgoing.setdefault(src, []).append((flow, tgt))
        incoming.setdefault(tgt, []).append((flow, src))

    # set real nodes (termasuk invisible yg akan kita tambahkan)
    real_nodes = {el.get("id") for el in elements.get("activities", []) + elements.get("events", [])}
    seen_edges = set()

    gateway_directions = {
        gid: classify_gateway_direction(gid, incoming, outgoing, gateway_map.get(gid))
        for gid in gateway_map.keys()
    }

    # masukkan invisible tasks sebagai activities agar diakui real node
    for inv in invisible_tasks:
        prev_node = inv.get("prev_node", {})  # ambil gateway/source sebelumnya
        pool_id = prev_node.get("pool_id")
        lane_id = prev_node.get("lane_id")
        pool_name = prev_node.get("pool_name")
        lane_name = prev_node.get("lane_name")

        pool_lane_props = ""
        if pool_id is not None:
            pool_lane_props += f"pool_id: {to_cypher_value(pool_id)}, "
        if lane_id is not None:
            pool_lane_props += f"lane_id: {to_cypher_value(lane_id)}, "
        if pool_name is not None:
            pool_lane_props += f"pool_name: {json.dumps(pool_name)}, "
        if lane_name is not None:
            pool_lane_props += f"lane_name: {json.dumps(lane_name)}, "

        cypher.append(
            f"CREATE (a:Activity {{"
            f"id: '{inv['id']}', "
            f"name: {json.dumps(inv.get('name', 'Invisible Task'))}, "
            f"type: '{inv['type']}', "
            f"{pool_lane_props}"
            f"process_id: '{process_id}'}});"
        )
        real_nodes.add(inv["id"])

    # --- fungsi recursive (TIDAK memblokir hanya berdasarkan gateway_chains) ---
    def find_real_targets(node_id, visited=None):
        if visited is None:
            visited = set()
        if node_id in visited:
            return []
        visited.add(node_id)

        next_targets = []
        for _, tgt in outgoing.get(node_id, []):
            if tgt in real_nodes:
                next_targets.append(tgt)
            elif tgt in outgoing:
                next_targets.extend(find_real_targets(tgt, visited))
        return next_targets

    def find_real_sources(node_id, visited=None):
        if visited is None:
            visited = set()
        if node_id in visited:
            return []
        visited.add(node_id)

        prev_sources = []
        for _, src in incoming.get(node_id, []):
            if src in real_nodes:
                prev_sources.append(src)
            elif src in incoming:
                prev_sources.extend(find_real_sources(src, visited))
        return prev_sources

    # --- BAGIAN PEMBENTUK EDGE --- 
    for flow in all_flows:
        src, tgt = flow.get("source"), flow.get("target")
        rel_name, gtype, direction = "SEQUENCE_FLOW", "", ""
        gateway_id = ""

        flow_type = (flow.get("flow_type") or flow.get("type") or "").lower()
        if "messageflow" in flow_type:
            rel_name = "MESSAGE_FLOW"

        if src in gateway_map:
            gtype = gateway_map[src]
            direction = gateway_directions.get(src, "SINGLE")
            rel_name = gateway_label(gtype, direction)
            gateway_id = src

        elif tgt in gateway_map:
            gtype = gateway_map[tgt]
            direction = gateway_directions.get(tgt, "SINGLE")
            rel_name = gateway_label(gtype, direction)
            gateway_id = tgt

        else:
            src_name = (flow.get("source_name") or "").lower()
            if "gateway" in src_name:
                gtype, direction = "gateway", "SINGLE"
                rel_name = gateway_label(gtype, direction)
                gateway_id = src
            else:
                gateway_id = ""

        rel_name = rel_name.upper().replace(" ", "_").replace("-", "_")
        if not rel_name.isidentifier():
            rel_name = "FLOW"

        props = (
            f"id: '{flow.get('id')}', name: '{flow.get('name','')}', "
            f"type: '{rel_name}', flow_type: '{flow.get('type','SEQUENCE')}', "
            f"gateway_type: '{gtype}', gateway_direction: '{direction}', "
            f"gateway_id: '{gateway_id}', "
            f"source_name: '{flow.get('source_name','')}', target_name: '{flow.get('target_name','')}', "
            f"source_pool: {json.dumps(flow.get('source_pool'))}, source_lane: {json.dumps(flow.get('source_lane'))}, "
            f"target_pool: {json.dumps(flow.get('target_pool'))}, target_lane: {json.dumps(flow.get('target_lane'))}, "
            f"source_pool_name: {json.dumps(flow.get('source_pool_name'))}, "
            f"source_lane_name: {json.dumps(flow.get('source_lane_name'))}, "
            f"target_pool_name: {json.dumps(flow.get('target_pool_name'))}, "
            f"target_lane_name: {json.dumps(flow.get('target_lane_name'))}, "
            f"process_id: '{process_id}'"
        )

        def add_edge(a_id, b_id):
            key = (a_id, b_id)
            if key not in seen_edges:
                seen_edges.add(key)
                cypher.append(
                    f"MATCH (a {{id: '{a_id}'}}) "
                    f"WITH a MATCH (b {{id: '{b_id}'}}) "
                    f"CREATE (a)-[:{rel_name} {{{props}}}]->(b);"
                )

        # event-based gateway special handling (tetap)
        if gtype == "eventbasedgateway":
            targets = [tgt] if tgt in real_nodes else find_real_targets(tgt)
            for real_tgt in targets:
                add_edge(src, real_tgt)
            continue

        # ---- KASUS: src real, tgt non-real (bisa gateway chain) ----
        if src in real_nodes and tgt not in real_nodes:
            # jika target adalah gateway yang termasuk gateway_chains:
            if tgt in gateway_chains:
                # izinkan bypass HANYA jika 'src' memang hanya punya satu outgoing (yaitu ke gateway ini)
                outs = outgoing.get(src, [])
                if len(outs) == 1:
                    # izinkan traversal normal
                    targets = find_real_targets(tgt)
                    for real_tgt in targets:
                        add_edge(src, real_tgt)
                else:
                    # jangan buat direct edges — biarkan Invisible Task meng-handle
                    pass
                continue

            # normal case
            targets = find_real_targets(tgt)
            for real_tgt in targets:
                add_edge(src, real_tgt)
            continue

        # ---- KASUS: src non-real, tgt real (bisa gateway chain di sebelah kiri) ----
        if src not in real_nodes and tgt in real_nodes:
            if src in gateway_chains:
                # izinkan reverse-bypass HANYA jika 'tgt' hanya punya satu incoming (yaitu dari gateway ini)
                ins = incoming.get(tgt, [])
                if len(ins) == 1:
                    sources = find_real_sources(src)
                    for real_src in sources:
                        add_edge(real_src, tgt)
                else:
                    # skip — biarkan Invisible Task meng-handle
                    pass
                continue

            # normal case
            sources = find_real_sources(src)
            for real_src in sources:
                add_edge(real_src, tgt)
            continue

        # ---- kedua tidak real ----
        if src not in real_nodes and tgt not in real_nodes:
            sources = find_real_sources(src)
            targets = find_real_targets(tgt)
            for s in sources:
                for t in targets:
                    add_edge(s, t)
            continue

        # ---- kedua real ----
        if src in real_nodes and tgt in real_nodes:
            add_edge(src, tgt)

    return cypher




from bpmn_neo4j.parsers.json_parser import load_json

class ParserFactory:
    @staticmethod
    def get_parser(file_path: str):
        if file_path.endswith(".json"):
            return JSONParser()
        raise ValueError("Unsupported file format")

class JSONParser:
    def parse(self, path):
        return load_json(path)



class AppException(Exception):
    """Base class untuk semua custom exception di aplikasi."""

    def __init__(self, message="Application error", code=500):
        super().__init__(message)
        self.message = message
        self.code = code

    def __str__(self):
        return f"{self.code}: {self.message}"



import json
import jsonschema
from jsonschema import validate
import uuid
import copy

def validate_schema(data, schema_path="validators/bpmn_schema.json", auto_fix=False):
    """
    Validasi struktur BPMN JSON dengan format baru (result.flowElements).
    """

    # Pastikan properti 'result' ada
    if "result" not in data:
        print("❌ The 'result' property is missing. This is required in BPMN structure.")
        data["result"] = {}

    # Pastikan struktur utama di dalam 'result' ada
    for key in ["flowElements", "messageFlows", "pools", "lanes"]:
        if key not in data["result"]:
            print(f"⚠️ Warning: The result key '{key}' is missing.")
            data["result"][key] = []

    # 🔍 Ambil flowElements
    flow_elements = data["result"].get("flowElements", [])

    # Deteksi dan perbaiki ID duplikat atau hilang
    if auto_fix:
        fix_missing_ids(data)
        fix_duplicate_ids(data)
    else:
        duplicates = check_duplicate_ids(data)
        if duplicates:
            print(f"❌ Duplicate IDs found: {duplicates}")

    # Deteksi siklus hanya pada flow bertipe "sequenceflow"
    sequence_flows = [f for f in flow_elements if f.get("type", "").lower() == "sequenceflow"]
    if detect_cycle(sequence_flows):
        print("⚠️ Warning: Circular reference detected in sequence flows.")
    else:
        print("✅ No circular dependencies in sequence flows.")

    # Validasi JSON terhadap schema resmi
    with open(schema_path, "r", encoding="utf-8") as f:
        schema = json.load(f)

    try:
        validate(instance=data, schema=schema)
        print("✅ JSON structure is valid against the new BPMN schema.")
        return data
    except jsonschema.exceptions.ValidationError as err:
        print(f"❌ Invalid structure: {err.message}")
        if auto_fix:
            print("🛠️ Attempting to fix the JSON structure automatically...")
            fixed_data = auto_fix_schema(data, schema)
            try:
                validate(instance=fixed_data, schema=schema)
                print("✅ JSON structure has been fixed successfully.")
                return fixed_data
            except jsonschema.exceptions.ValidationError as err2:
                print(f"❌ Structure fixing failed: {err2.message}")
                return fixed_data
        else:
            return data


# 🔁 Cek ID duplikat di semua flowElements
def check_duplicate_ids(data):
    id_set = set()
    duplicates = []

    for el in data["result"].get("flowElements", []):
        iid = el.get("id")
        if iid:
            if iid in id_set:
                duplicates.append(iid)
            id_set.add(iid)
    return duplicates


# ✅ Perbaiki ID duplikat dengan suffix
def fix_duplicate_ids(data):
    print("🛠️ Checking and fixing duplicate IDs...")
    id_count = {}
    for el in data["result"].get("flowElements", []):
        iid = el.get("id")
        if not iid:
            continue
        if iid in id_count:
            id_count[iid] += 1
            new_id = f"{iid}_{id_count[iid]}"
            print(f"⚠️ Duplicate ID '{iid}' found. Renaming to '{new_id}'.")
            el["id"] = new_id
        else:
            id_count[iid] = 0


# ✅ Beri ID default jika hilang
def fix_missing_ids(data):
    print("🛠️ Checking for missing IDs...")
    for el in data["result"].get("flowElements", []):
        if "id" not in el or not el["id"]:
            new_id = f"element_{uuid.uuid4().hex[:6]}"
            print(f"⚠️ Missing ID detected. Assigning ID: {new_id}")
            el["id"] = new_id


# 🧠 Auto-fix berdasarkan JSON schema (rekursif)
def auto_fix_schema(data, schema):
    def fix_object(obj, schema_obj, path="root"):
        if not isinstance(obj, dict):
            return

        required_props = schema_obj.get("required", [])
        properties = schema_obj.get("properties", {})

        # Tambahkan properti wajib yang hilang
        for key in required_props:
            if key not in obj:
                prop_schema = properties.get(key, {})
                default_value = generate_default_value(key, prop_schema, path)
                obj[key] = default_value
                print(f"🛠️ Auto-added missing '{key}' at {path}: {default_value}")

        # Rekursif untuk nested objek
        for key, val in obj.items():
            if key in properties:
                prop_schema = properties[key]
                if isinstance(val, dict) and prop_schema.get("type") == "object":
                    fix_object(val, prop_schema, f"{path}.{key}")
                elif isinstance(val, list) and prop_schema.get("type") == "array":
                    item_schema = prop_schema.get("items", {})
                    for idx, item in enumerate(val):
                        if isinstance(item, dict):
                            fix_object(item, item_schema, f"{path}.{key}[{idx}]")

    def generate_default_value(key, prop_schema, path=""):
        if prop_schema.get("enum"):
            return prop_schema["enum"][0]
        if prop_schema.get("type") == "string":
            if key == "id":
                return f"{path.replace('.', '_')}_{uuid.uuid4().hex[:6]}"
            return f"default_{key}"
        if prop_schema.get("type") == "array":
            return []
        if prop_schema.get("type") == "object":
            return {}
        return None

    fixed_data = copy.deepcopy(data)
    fix_object(fixed_data, schema, path="root")
    return fixed_data


# 🔁 Deteksi siklus hanya untuk sequenceFlow
def detect_cycle(flows):
    from collections import defaultdict, deque

    graph = defaultdict(list)
    indegree = defaultdict(int)
    nodes = set()

    for f in flows:
        src = f.get("source")
        tgt = f.get("target")
        if src and tgt:
            graph[src].append(tgt)
            indegree[tgt] += 1
            nodes.update([src, tgt])

    queue = deque([n for n in nodes if indegree[n] == 0])
    visited = 0

    while queue:
        current = queue.popleft()
        visited += 1
        for neighbor in graph[current]:
            indegree[neighbor] -= 1
            if indegree[neighbor] == 0:
                queue.append(neighbor)

    return visited != len(nodes)
